/*
 *  PyGiNaC Copyright (C) 2019 Vladimir V. Kisil
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <ginac/ginac.h>

#include <boost/python/class.hpp>
#include <boost/python/operators.hpp>
#include <boost/python/implicit.hpp>
#include <boost/python/extract.hpp>
#include <boost/python/overloads.hpp>
#include <boost/python/def.hpp>
#include <boost/python/dict.hpp>
#include <boost/python/make_constructor.hpp>

GiNaC::symtab
dict2symtab( boost::python::dict syms_)
{
    using namespace boost::python;
    object iter = syms_.attr( "items")().attr("__iter__")();
    GiNaC::symtab table;

    while (true) {
        try {
            object item = iter.attr("__next__")();
            table[ extract<std::string>(item[0])] = extract<GiNaC::ex>(item[1]);
        }
        catch (boost::python::error_already_set& ) {
            PyErr_Clear();
            break;
        }
    }

	return table;
}

static boost::shared_ptr<GiNaC::parser>
init_parser( boost::python::dict syms_, const bool strict_)
{
    return boost::shared_ptr<GiNaC::parser>(new GiNaC::parser(dict2symtab(syms_), strict_));
}

GiNaC::ex
do_parser(GiNaC::parser & reader, const std::string& input) {
    return reader(input);
}

void
set_strict(GiNaC::parser & reader, const bool strict_) {
    reader.strict = strict_;
}

void
set_syms(GiNaC::parser & reader, boost::python::dict syms_) {
    reader.get_syms() = dict2symtab(syms_);
}

namespace pyginac {

// Implicit conversion from GiNaC::symtab -> dict { ex : ex }
namespace {
struct symtab_to_python
{
    static PyObject* convert( const GiNaC::symtab& obj)
    {
        boost::python::dict ret;
        for (const auto& x: obj)
            ret[x.first] = x.second;
        Py_INCREF(ret.ptr());
        return ret.ptr();
    }
};
} // !namespace anonymous
    
void
wrap_symtab(void)
{
    boost::python::to_python_converter<GiNaC::symtab, symtab_to_python>();
}
    
void
wrap_parser(void)
{
    using namespace boost::python;
    using GiNaC::parser;

    // Difficult overloads
    GiNaC::symtab (GiNaC::parser::*get_syms_const)() const = & GiNaC::parser::get_syms;
    
    class_<parser>( "parser", "Recursive descent parser for GiNaC expressions.",
                    init<>())
        .def("__init__", make_constructor(init_parser,  default_call_policies(),
                (boost::python::arg("syms_")=boost::python::dict(), boost::python::arg("strict_") = false)), "Inititialise a parser for GiNaC expressions")
        .def( "__call__", do_parser)
        .def( "get_syms", get_syms_const, "Report the symbol table used by parser")
        .def( "set_syms", set_syms, "Set the parser's symbol table from a Python dictionary")
        .def( "set_strict", set_strict, "Set strict property of the parser on accepting unknown symbols")
        ;
        
}

} // !namespace pyginac
